"""Tests for the 'async' backend."""

import pytest

from pathlib import Path
from typing import Any

import dawgz

from dawgz.schedulers import CyclicDependencyGraphError

########
# Jobs #
########


@dawgz.job
def identity(x: Any) -> Any:
    return x


@dawgz.job
def fail() -> None:
    raise RuntimeError("intentional failure")


@dawgz.job
def mutate_list(items: list) -> list:
    items.append("mutated")
    return items


############
# Fixtures #
############


@pytest.fixture(autouse=True)
def redirect_dawgz_dir(tmp_path: Path) -> None:
    dawgz.set_dawgz_dir(tmp_path / ".dawgz")


@pytest.fixture(params=[None, 4], ids=["threads", "processes"])
def pools(request: pytest.FixtureRequest) -> int | None:
    return request.param


#########
# Tests #
#########


def test_single_job(pools: int | None) -> None:
    job = identity("x")
    scheduler = dawgz.schedule(job, backend="async", quiet=True, pools=pools)
    assert scheduler.results[job] == "x"
    assert job not in scheduler.traces


def test_failing_job(pools: int | None) -> None:
    job = fail()
    scheduler = dawgz.schedule(job, backend="async", quiet=True, pools=pools)
    assert "RuntimeError" in scheduler.traces[job]
    assert job not in scheduler.results


def test_mutation(pools: int | None) -> None:
    items = ["a", "b", "c"]
    job = mutate_list(items)
    items.append("d")
    scheduler = dawgz.schedule(job, backend="async", quiet=True, pools=pools)
    assert "d" not in scheduler.results[job]
    assert "mutated" in scheduler.results[job]
    assert "mutated" not in items


def test_linear_chain(pools: int | None) -> None:
    a, b, c = "abc"
    a_job = identity(a)
    b_job = identity(b).after(a_job)
    c_job = identity(c).after(b_job)
    scheduler = dawgz.schedule(c_job, backend="async", quiet=True, pools=pools)
    assert scheduler.results[a_job] == a
    assert scheduler.results[b_job] == b
    assert scheduler.results[c_job] == c


def test_fan_out(pools: int | None) -> None:
    a, b, c, d = "abcd"
    a_job = identity(a)
    b_job = identity(b).after(a_job)
    c_job = identity(c).after(a_job)
    d_job = identity(d).after(a_job)
    scheduler = dawgz.schedule(b_job, c_job, d_job, backend="async", quiet=True, pools=pools)
    assert scheduler.results[a_job] == a
    assert scheduler.results[b_job] == b
    assert scheduler.results[c_job] == c
    assert scheduler.results[d_job] == d


def test_fan_in(pools: int | None) -> None:
    a, b, c, d = "abcd"
    a_job = identity(a)
    b_job = identity(b)
    c_job = identity(c)
    d_job = identity(d).after(a_job, b_job, c_job)
    scheduler = dawgz.schedule(d_job, backend="async", quiet=True, pools=pools)
    assert scheduler.results[a_job] == a
    assert scheduler.results[b_job] == b
    assert scheduler.results[c_job] == c
    assert scheduler.results[d_job] == d


def test_diamond(pools: int | None) -> None:
    a, b, c, d = "abcd"
    a_job = identity(a)
    b_job = identity(b).after(a_job)
    c_job = identity(c).after(a_job)
    d_job = identity(d).after(b_job, c_job)
    scheduler = dawgz.schedule(d_job, backend="async", quiet=True, pools=pools)
    assert scheduler.results[a_job] == a
    assert scheduler.results[b_job] == b
    assert scheduler.results[c_job] == c
    assert scheduler.results[d_job] == d


def test_cyclic_dependency(pools: int | None) -> None:
    a_job = identity(None)
    b_job = identity(None)
    a_job.after(b_job)
    b_job.after(a_job)
    with pytest.raises(CyclicDependencyGraphError):
        dawgz.schedule(a_job, backend="async", quiet=True, pools=pools)


def test_blocked_by_failed_dep(pools: int | None) -> None:
    a_job = fail()
    b_job = identity(None).after(a_job)
    scheduler = dawgz.schedule(b_job, backend="async", quiet=True, pools=pools)
    assert "RuntimeError" in scheduler.traces[a_job]
    assert "DependencyNeverSatisfiedError" in scheduler.traces[b_job]


def test_triggered_by_failed_dep(pools: int | None) -> None:
    a_job = fail()
    b_job = identity("b").after(a_job, status="failure")
    scheduler = dawgz.schedule(b_job, backend="async", quiet=True, pools=pools)
    assert "RuntimeError" in scheduler.traces[a_job]
    assert scheduler.results[b_job] == "b"


def test_waitfor_any_one_succeeds(pools: int | None) -> None:
    a_job = fail()
    b_job = identity("b")
    c_job = identity("c").after(a_job, b_job).waitfor("any")
    scheduler = dawgz.schedule(c_job, backend="async", quiet=True, pools=pools)
    assert "RuntimeError" in scheduler.traces[a_job]
    assert scheduler.results[b_job] == "b"
    assert scheduler.results[c_job] == "c"


def test_waitfor_any_all_fail(pools: int | None) -> None:
    a_job = fail()
    b_job = fail()
    c_job = identity("c").after(a_job, b_job).waitfor("any")
    scheduler = dawgz.schedule(c_job, backend="async", quiet=True, pools=pools)
    assert "RuntimeError" in scheduler.traces[a_job]
    assert "RuntimeError" in scheduler.traces[b_job]
    assert "DependencyNeverSatisfiedError" in scheduler.traces[c_job]


def test_mark_success(pools: int | None) -> None:
    a_job = fail().mark("success")
    b_job = identity("b").after(a_job)
    scheduler = dawgz.schedule(b_job, backend="async", quiet=True, pools=pools)
    assert a_job not in scheduler.results
    assert a_job not in scheduler.traces
    assert scheduler.results[b_job] == "b"


def test_mark_failure(pools: int | None) -> None:
    a_job = identity("a").mark("failure")
    b_job = identity("b").after(a_job, status="failure")
    scheduler = dawgz.schedule(b_job, backend="async", quiet=True, pools=pools)
    assert a_job not in scheduler.results
    assert a_job not in scheduler.traces
    assert scheduler.results[b_job] == "b"


def test_job_array(pools: int | None) -> None:
    xs = "abc"
    array = dawgz.array(*map(identity, xs))
    scheduler = dawgz.schedule(array, backend="async", quiet=True, pools=pools)
    assert all(scheduler.results[array][i] == x for i, x in enumerate(xs))


def test_job_array_with_dep(tmp_path: Path, pools: int | None) -> None:
    a_job = identity("a")
    xs = "abc"
    array = dawgz.array(*map(identity, xs)).after(a_job)
    scheduler = dawgz.schedule(array, backend="async", quiet=True, pools=pools)
    assert scheduler.results[a_job] == "a"
    assert all(scheduler.results[array][i] == x for i, x in enumerate(xs))



r"""Workflow graph components"""

from __future__ import annotations

from collections.abc import Callable, Iterator, Sequence
from functools import partial
from rich.pretty import pretty_repr
from textwrap import indent
from typing import (
    Any,
    Literal,
    TypeVar,
)

from .utils import pickle


class Node:
    r"""Abstract graph node"""

    def __init__(self) -> None:
        self.children = {}
        self.parents = {}

    def add_child(self, node: Node, edge: Any | None = None) -> None:
        self.children[node] = edge
        node.parents[self] = edge

    def add_parent(self, node: Node, edge: Any | None = None) -> None:
        node.add_child(self, edge)

    def rm_child(self, node: Node) -> None:
        del self.children[node]
        del node.parents[self]

    def rm_parent(self, node: Node) -> None:
        node.rm_child(self)


class Job(Node):
    r"""Job node."""

    def __init__(
        self,
        fun: Callable | None,
        args: Sequence[Any] = (),
        kwargs: dict[str, Any] = {},  # noqa: B006
        *,
        name: str | None = None,
        shell: str = "/bin/bash",
        interpreter: str = "python",
        env: list[str] | None = None,
        settings: dict[str, Any] = {},  # noqa: B006
    ) -> None:
        super().__init__()

        if fun is None:
            self.exe = None
        else:
            self.exe = pickle.dumps(partial(fun, *args, **kwargs))

        # String repr
        if name is None:
            name = getattr(fun, "__name__", None)

        assert isinstance(name, str) and name.replace("_", "").isalnum(), (
            f"function name can only contain underscore and alphanumeric characters, got '{name}'"
        )

        def prepr(x: object) -> str:
            return pretty_repr(
                x,
                indent_size=2,
                max_depth=2,
                max_width=48,
                max_string=24,
            ).strip("\n")

        self.name = name
        self.args_repr = [prepr(a) for a in args] + [f"{k}=" + prepr(v) for k, v in kwargs.items()]

        # Settings
        self.shell = shell
        self.interpreter = interpreter
        self.env = env
        self.settings = settings

        # Status
        self.status: str = "pending"

        # Dependencies
        self.wait_mode: str = "all"
        self.satisfied: dict[Job, str] = {}
        self.unsatisfied: dict[Job, str] = {}

    def __repr__(self) -> str:
        prepr = f"{self.name}(" + ", ".join(self.args_repr) + ")"

        if "\n" in prepr or len(prepr) > 48:
            prepr = f"{self.name}(\n" + indent(",\n".join(self.args_repr), "  ") + "\n)"

        return prepr

    def __getstate__(self) -> dict:
        state = self.__dict__.copy()
        state.pop("exe", None)
        return state

    def mark(self, status: Literal["success", "failure", "cancelled", "pending"]) -> Job:
        r"""Sets the completion status of a job.

        Arguments:
            status: The completion status. The default status is `"pending"`.
        """
        assert status in ["success", "failure", "cancelled", "pending"]
        self.status = status
        return self

    @property
    def dependencies(self) -> dict[Job, str]:
        return self.parents

    def after(self, *deps: Job, status: Literal["success", "failure", "any"] = "success") -> Job:
        r"""Adds dependencies to a job.

        Arguments:
            deps: A set of job dependencies.
            status: The desired dependency status.
        """
        assert status in ["success", "failure", "any"]
        for dep in deps:
            self.add_parent(dep, status)
        return self

    def detach(self, *deps: Job) -> None:
        for dep in deps:
            self.rm_parent(dep)

    def waitfor(self, mode: Literal["all", "any"]) -> Job:
        r"""Sets the waiting mode of a job.

        Arguments:
            mode: The dependency waiting mode. The default mode is `"all"`.
        """
        assert mode in ["all", "any"]
        self.wait_mode = mode
        return self

    @property
    def satisfy_status(self) -> Literal["ready", "never", "wait"]:
        if self.wait_mode == "all" and self.unsatisfied:
            return "never"
        elif (
            self.wait_mode == "all"
            and not self.dependencies
            or self.wait_mode == "any"
            and self.satisfied
        ):
            return "ready"
        elif self.wait_mode == "any" and not self.dependencies:
            return "never"
        else:
            return "wait"


class JobArray(Job):
    def __init__(self, *jobs: Job, throttle: int | None = None) -> None:
        self.array = jobs
        self.throttle = throttle

        assert len(self.array) >= 1, "array should contain at least one job"
        assert len(self.array) == len(set(self.array)), "array should not contain duplicates"

        super().__init__(
            fun=None,
            name=self.array[0].name,
            interpreter=self.array[0].interpreter,
            env=self.array[0].env,
            settings=self.array[0].settings,
        )

        for job in self.array:
            assert not job.parents, "jobs in an array should not have dependencies"
            assert not job.children, "jobs in an array should not have dependents"

        for key in ("name", "shell", "interpreter", "env", "settings"):
            for job in self.array:
                assert getattr(job, key) == getattr(self, key), (
                    f"all jobs in an array should have the same {key}"
                )

    def __len__(self) -> int:
        return len(self.array)

    def __getitem__(self, i: int) -> Job:
        return self.array[i]

    def __repr__(self) -> str:
        if self.throttle is None:
            range = f"0-{len(self) - 1}"
        else:
            range = f"0-{len(self) - 1}%{self.throttle}"

        return f"{self.name}[{range}]"


N = TypeVar("N", bound=Node)


def dfs(*nodes: N, backward: bool = False) -> Iterator[N]:
    queue = list(nodes)
    visited = set()

    while queue:
        node = queue.pop()

        if node in visited:
            continue
        else:
            yield node

        queue.extend(node.parents if backward else node.children)
        visited.add(node)


def leafs(*nodes: N) -> set[N]:
    return {node for node in dfs(*nodes, backward=False) if not node.children}


def roots(*nodes: N) -> set[N]:
    return {node for node in dfs(*nodes, backward=True) if not node.parents}


def cycles(*nodes: Node, backward: bool = False) -> Iterator[list[Node]]:
    queue = [list(nodes)]
    path = []
    pathset = set()
    visited = set()

    while queue:
        branch = queue[-1]

        if not branch:
            if not path:
                break

            queue.pop()
            pathset.remove(path.pop())
            continue

        node = branch.pop()

        if node in visited:
            if node in pathset:
                yield path + [node]
            continue

        queue.append(list(node.parents if backward else node.children))
        path.append(node)
        pathset.add(node)
        visited.add(node)


def prune(*jobs: Job) -> list[Job]:
    for job in dfs(*jobs, backward=True):
        if job.status != "pending":
            job.detach(*job.dependencies)

        for dep, status in job.dependencies.items():
            if dep.status == "pending":
                pass
            elif dep.status == "cancelled":
                job.unsatisfied[dep] = status
            elif status == "any" or dep.status == status:
                job.satisfied[dep] = status
            else:
                job.unsatisfied[dep] = status

        job.detach(*job.satisfied, *job.unsatisfied)

    return list(dict.fromkeys(job for job in jobs if job.status == "pending"))



r"""Scheduling backends"""

from __future__ import annotations

import asyncio
import concurrent.futures as cf
import csv
import random
import shutil
import subprocess

from abc import ABC, abstractmethod
from collections.abc import Iterator
from contextlib import contextmanager
from datetime import datetime
from functools import cache
from pathlib import Path
from tabulate import tabulate
from typing import Any

from .constants import get_dawgz_dir
from .utils import cat, future, human_uuid, pickle, runpickle, slugify, trace
from .workflow import Job, JobArray, cycles, prune


class Scheduler(ABC):
    r"""Abstract workflow scheduler."""

    backend: str = None

    def __init__(self, name: str | None = None) -> None:
        r"""
        Arguments:
            name: The name of the workflow.
            settings: A dictionnary of settings.
            kwargs: Keyword arguments added to `settings`.
        """

        super().__init__()

        self.name = name
        self.date = datetime.now().replace(microsecond=0)
        self.uid = human_uuid()

        self.path = get_dawgz_dir() / self.uid
        self.path.mkdir(parents=True)

        # Jobs
        self.order = {}
        self.results = {}
        self.traces = {}

    def dump(self) -> None:
        with open(self.path / "dump.pkl", "wb") as f:
            pickle.dump(self, f)

        with open(self.path.parent / "workflows.csv", mode="a", newline="") as f:
            csv.writer(f).writerow((
                self.name,
                self.uid,
                self.date,
                self.backend,
                len(self.order),
                len(self.traces),
            ))

    @staticmethod
    def load(path: Path) -> Scheduler:
        with open(path / "dump.pkl", "rb") as f:
            return pickle.load(f)

    def tag(self, job: Job) -> str:
        if job in self.order:
            i = self.order[job]
        else:
            i = self.order[job] = len(self.order)

        return f"{i:04d}_{slugify(job.name)}"

    def state(self, job: Job, i: int | None = None) -> str:
        if job in self.traces:
            return "FAILED"
        else:
            return "COMPLETED"

    def output(self, job: Job, i: int | None = None) -> Any:
        if isinstance(job, JobArray):
            return self.results[job][i]
        else:
            return self.results[job]

    def report(self, job: Job | None = None, i: int | None = None) -> str:
        if job is None:
            headers = ("Job", "State")
            rows = [(str(job), self.state(job)) for job in self.order]

            return tabulate(rows, headers, showindex=True, maxcolwidths=[None, 48, 16])
        else:
            headers = ("Job", "State", "Output")

            if job in self.traces:
                index = [None]
                rows = [(str(job), self.state(job), self.traces[job])]
            elif isinstance(job, JobArray):
                index = list(range(len(job))) if i is None else [i]
                rows = [(str(job[i]), self.state(job, i), self.output(job, i)) for i in index]
            else:
                index = [None]
                rows = [(str(job), self.state(job), self.output(job))]

            terminal_width = shutil.get_terminal_size((0, 0)).columns

            rows = [
                (
                    repr,
                    state,
                    cat("" if output is None else str(output), width=terminal_width - 64 - 3 * 2),
                )
                for repr, state, output in rows
            ]

            return tabulate(rows, headers, showindex=index, maxcolwidths=[None, 48, 16, None])

    def cancel(self, job: Job | None = None, i: int | None = None) -> str:
        raise NotImplementedError(f"'cancel' is not implemented for the {self.backend} backend.")

    @contextmanager
    def context(self) -> Iterator[None]:
        try:
            yield None
        finally:
            pass

    def __call__(self, *jobs: Job) -> None:
        for cycle in cycles(*jobs, backward=True):
            raise CyclicDependencyGraphError(" <- ".join(map(str, cycle)))

        jobs = prune(*dict.fromkeys(jobs))

        with self.context():
            asyncio.run(self.wait(*jobs))

    async def wait(self, *jobs: Job) -> None:
        self.tasks: dict[Job, asyncio.Task] = {}

        if jobs:
            await asyncio.wait(map(asyncio.create_task, map(self.submit, jobs)))
            await asyncio.wait(self.tasks.values())

        del self.tasks

    async def submit(self, job: Job) -> Any:
        if job in self.tasks:
            task = self.tasks[job]
        else:
            task = self.tasks[job] = future(self._submit(job), return_exceptions=True)

        result = await task

        if isinstance(result, Exception):
            self.traces[job] = trace(result)
        else:
            self.results[job] = result

        return result

    async def _submit(self, job: Job) -> Any:
        try:
            if job.satisfy_status == "ready":
                asyncio.ensure_future(self.satisfy(job))
            elif job.satisfy_status == "wait":
                await self.satisfy(job)
            elif job.satisfy_status == "never":
                raise DependencyNeverSatisfiedError(str(job))
            else:
                raise ValueError(f"Unknown status '{job.satisfy_status}'.")
        finally:
            self.tag(job)

        return await future(self.exec(job), return_exceptions=True)

    @abstractmethod
    async def satisfy(self, job: Job) -> None:
        pass

    @abstractmethod
    async def exec(self, job: Job) -> Any:
        pass


class AsyncScheduler(Scheduler):
    r"""Asynchronous scheduler.

    Jobs are executed asynchronously. A job is launched as soon as its dependencies are
    satisfied.
    """

    backend: str = "async"

    def __init__(self, name: str | None = None, pools: int | None = None) -> None:
        r"""
        Arguments:
            name: The name of the workflow.
            pools: The number of processing pools. If `None`, use threads instead.
        """

        super().__init__(name=name)

        self.pools = pools

    @contextmanager
    def context(self) -> Iterator[None]:
        if self.pools is None:
            self.executor = cf.ThreadPoolExecutor()
        else:
            self.executor = cf.ProcessPoolExecutor(self.pools)

        try:
            yield None
        finally:
            del self.executor

    async def satisfy(self, job: Job) -> None:
        pending = [
            asyncio.gather(self.submit(dep), future(status))
            for dep, status in job.dependencies.items()
        ]

        while pending:
            done, pending = await asyncio.wait(pending, return_when=asyncio.FIRST_COMPLETED)

            for task in done:
                result, status = task.result()

                if isinstance(result, JobFailedError) and status != "success":
                    result = None
                elif not isinstance(result, JobFailedError) and status == "failure":
                    result = JobNotFailedError(str(job))

                if isinstance(result, Exception):
                    if job.wait_mode == "all":
                        raise DependencyNeverSatisfiedError(str(job)) from result
                elif job.wait_mode == "any":
                    break
            else:
                continue
            break
        else:
            if job.wait_mode == "any":
                raise DependencyNeverSatisfiedError(str(job))

    async def exec(self, job: Job) -> Any:
        loop = asyncio.get_running_loop()

        try:
            if isinstance(job, JobArray):
                futures = [
                    loop.run_in_executor(self.executor, runpickle, job[i].exe)
                    for i in range(len(job))
                ]

                results = await asyncio.gather(*futures, return_exceptions=True)

                for result in results:
                    if isinstance(result, Exception):
                        raise result

                return results
            else:
                return await loop.run_in_executor(self.executor, runpickle, job.exe)
        except Exception as e:
            raise JobFailedError(str(job)) from e


class DummyScheduler(AsyncScheduler):
    r"""Dummy asynchronous scheduler.

    Jobs are scheduled asynchronously, but instead of executing them, their name is
    printed before and after a short (random) sleep time. Useful for debugging.
    """

    backend: str = "dummy"

    async def exec(self, job: Job) -> None:
        print(f"START {job}")
        await asyncio.sleep(random.random())
        print(f"END   {job}")


class SlurmScheduler(Scheduler):
    r"""Slurm scheduler.

    Jobs are submitted to the Slurm queue. Resources are allocated by the Slurm manager
    according to the job and scheduler settings. Job settings have precendence over
    scheduler settings.

    Most settings (e.g. `account`, `export`, `partition`) are passed directly to
    `sbatch`. A few settings (e.g. `cpus`, `gpus`, `ram`) are translated into their
    `sbatch` equivalents.
    """

    backend: str = "slurm"
    translate: dict[str, str] = {  # noqa: RUF012
        "tasks": "ntasks",
        "tasks_per_node": "ntasks-per-node",
        "cpus": "cpus-per-task",
        "gpus": "gpus-per-task",
        "ram": "mem",
        "memory": "mem",
        "timelimit": "time",
        "timeout": "time",
    }

    def __init__(self, name: str | None = None) -> None:
        r"""
        Arguments:
            name: The name of the workflow.
        """

        super().__init__(name=name)

    @staticmethod
    @cache
    def sacct(jobid: str) -> dict[str, str]:
        text = subprocess.run(
            ["sacct", "-j", jobid, "-o", "JobID,State", "-n", "-P", "-X"],
            capture_output=True,
            check=True,
            text=True,
        ).stdout.strip("\n")

        return dict(line.split("|") for line in text.splitlines())

    def state(self, job: Job, i: int | None = None) -> str | None:
        if job in self.traces:
            return "CANCELLED"

        jobid = self.results[job]
        table = self.sacct(jobid)

        if isinstance(job, JobArray):
            if i is None:
                return ", ".join(sorted(set(table.values())))
            else:
                return table.get(f"{jobid}_{i}", None)
        else:
            return table.get(jobid, None)

    def output(self, job: Job, i: int | None = None) -> str | None:
        tag = self.tag(job)

        if isinstance(job, JobArray):
            logfile = self.path / f"{tag}_{i}.log"
        else:
            logfile = self.path / f"{tag}.log"

        if logfile.exists():
            with open(logfile, newline="", errors="replace") as f:
                return f.read()
        else:
            return None

    def report(self, job: Job | None = None, i: int | None = None) -> str:
        if job is None:
            headers = ("Job", "ID", "State")
            rows = []

            for job in self.order:
                if job in self.traces:
                    jobid = None
                else:
                    jobid = self.results[job]

                rows.append((str(job), jobid, self.state(job)))

            return tabulate(rows, headers, showindex=True, maxcolwidths=[None, 48, 16, 16])
        else:
            return super().report(job, i)

    def cancel(self, job: Job | None = None, i: int | None = None) -> str:
        if job is None:
            jobids = list(self.results.values())
        else:
            jobid = self.results[job]
            if i is not None:  # fails if job is not array
                jobid = f"{jobid}_{i}"
            jobids = [jobid]

        return subprocess.run(
            ["scancel", "-v", *jobids],
            capture_output=True,
            check=True,
            text=True,
        ).stderr.strip("\n")

    async def satisfy(self, job: Job) -> str:
        results = await asyncio.gather(*map(self.submit, job.dependencies))

        for result in results:
            if isinstance(result, Exception):
                raise DependencyNeverSatisfiedError(str(job)) from result

    async def exec(self, job: Job) -> str:
        tag = self.tag(job)
        logfile = self.path / (f"{tag}_%a.log" if isinstance(job, JobArray) else f"{tag}.log")
        pklfile = self.path / f"{tag}.pkl"
        pyfile = self.path / f"{tag}.py"
        shfile = self.path / f"{tag}.sh"

        # Submission script
        lines = [
            f"#!{job.shell}",
            "#",
            f"#SBATCH --job-name={tag}",
        ]

        if isinstance(job, JobArray):
            if job.throttle is None:
                lines.append(f"#SBATCH --array=0-{len(job) - 1}")
            else:
                lines.append(f"#SBATCH --array=0-{len(job) - 1}%{job.throttle}")

        lines.append(f"#SBATCH --output={logfile}")
        lines.append("#")

        ## Settings
        settings = {self.translate.get(k, k).replace("_", "-"): v for k, v in job.settings.items()}

        assert "clusters" not in settings, "multi-cluster jobs not supported"

        if "ntasks" not in settings:
            settings.setdefault("nodes", 1)
            settings.setdefault("ntasks-per-node", 1)

        for key, value in sorted(settings.items()):
            if isinstance(value, bool) and value:
                lines.append(f"#SBATCH --{key}")
            else:
                lines.append(f"#SBATCH --{key}={value}")

        ## Dependencies
        sep = "?" if job.wait_mode == "any" else ","
        after = {
            "success": "afterok",
            "failure": "afternotok",
            "any": "afterany",
        }

        deps = [
            f"{after[status]}:{await self.submit(dep)}" for dep, status in job.dependencies.items()
        ]

        if deps:
            lines.append("#")
            lines.append("#SBATCH --dependency=" + sep.join(deps))

        lines.append("")

        ## Environment
        if job.env:
            lines.extend(job.env)
            lines.append("")

        ## Pickle job
        if isinstance(job, JobArray):
            pklfile = str(pklfile).replace(".pkl", "_{}.pkl")

            for i in range(len(job)):
                with open(pklfile.format(i), mode="wb") as f:
                    f.write(job[i].exe)

            with open(pyfile, mode="w") as f:
                f.write(
                    "\n".join([
                        "#!/usr/bin/env python",
                        "import os",
                        "import pickle",
                        "i = os.environ['SLURM_ARRAY_TASK_ID']",
                        f"with open('{pklfile}'.format(i), 'rb') as f:",
                        "    pickle.load(f)()",
                    ])
                )
        else:
            with open(pklfile, mode="wb") as f:
                f.write(job.exe)

            with open(pyfile, mode="w") as f:
                f.write(
                    "\n".join([
                        "#!/usr/bin/env python",
                        "import pickle",
                        f"with open('{pklfile}', 'rb') as f:",
                        "    pickle.load(f)()",
                    ])
                )

        lines.append(f"srun {job.interpreter} {pyfile}")
        lines.append("")

        ## Save
        with open(shfile, mode="w") as f:
            f.write("\n".join(lines))

        # Submit script
        try:
            text = subprocess.run(
                ["sbatch", "--parsable", str(shfile)],
                capture_output=True,
                check=True,
                text=True,
            ).stdout

            jobid, *_ = text.strip("\n").split(";")  # ignore cluster name

            return jobid
        except Exception as e:
            if isinstance(e, subprocess.CalledProcessError):
                e = subprocess.SubprocessError(e.stderr.strip("\n"))

            raise JobSubmissionError(str(job)) from e


class CyclicDependencyGraphError(Exception):
    pass


class DependencyNeverSatisfiedError(Exception):
    pass


class JobFailedError(Exception):
    pass


class JobNotFailedError(Exception):
    pass


class JobSubmissionError(Exception):
    pass



"""Tests for the dawgz CLI."""

import pytest
import sys

from pathlib import Path

import dawgz

from dawgz.__main__ import main

########
# Jobs #
########


@dawgz.job
def noop() -> int:
    return 42


@dawgz.job
def failing(arg: str, kwarg: str) -> None:
    raise RuntimeError("intentional failure")


############
# Fixtures #
############


@pytest.fixture(autouse=True)
def redirect_dawgz_dir(tmp_path: Path) -> None:
    dawgz.set_dawgz_dir(tmp_path / ".dawgz")


@pytest.fixture()
def dummy_workflow() -> dawgz.Scheduler:
    return dawgz.schedule(
        noop(), failing("a", kwarg="k"), name="dummy", backend="async", quiet=True
    )


#########
# Tests #
#########


def test_main_no_workflows(capsys: pytest.CaptureFixture, monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(sys, "argv", ["dawgz"])
    main()
    out = capsys.readouterr().out
    assert "Name" in out
    assert "Backend" in out


def test_main_workflows(
    dummy_workflow: dawgz.Scheduler, capsys: pytest.CaptureFixture, monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.setattr(sys, "argv", ["dawgz"])
    main()
    out = capsys.readouterr().out
    assert "dummy" in out
    assert "async" in out


def test_main_workflow(
    dummy_workflow: dawgz.Scheduler, capsys: pytest.CaptureFixture, monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.setattr(sys, "argv", ["dawgz", "0"])
    main()
    out = capsys.readouterr().out
    assert "noop()" in out
    assert "COMPLETED" in out
    assert "failing('a', kwarg='k')" in out
    assert "FAILED" in out


def test_main_job(
    dummy_workflow: dawgz.Scheduler, capsys: pytest.CaptureFixture, monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.setattr(sys, "argv", ["dawgz", "-1", "0"])
    main()
    out = capsys.readouterr().out
    assert "noop()" in out
    assert "COMPLETED" in out
    assert "42" in out


def test_main_job_failing(
    dummy_workflow: dawgz.Scheduler, capsys: pytest.CaptureFixture, monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.setattr(sys, "argv", ["dawgz", "-1", "-1"])
    main()
    out = capsys.readouterr().out
    assert "failing('a', kwarg='k')" in out
    assert "FAILED" in out
    assert "RuntimeError" in out


def test_main_invalid_workflow_index(
    dummy_workflow: dawgz.Scheduler, monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.setattr(sys, "argv", ["dawgz", "99"])
    with pytest.raises(IndexError):
        main()


def test_main_invalid_job_index(
    dummy_workflow: dawgz.Scheduler, monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.setattr(sys, "argv", ["dawgz", "0", "99"])
    with pytest.raises(IndexError):
        main()



"""Tests for the 'slurm' backend."""

import os
import pytest
import re
import subprocess
import sys

from collections.abc import Generator
from pathlib import Path
from unittest.mock import MagicMock, patch

import dawgz

from dawgz.schedulers import SlurmScheduler

############
# Fixtures #
############


@pytest.fixture(autouse=True)
def redirect_dawgz_dir(tmp_path: Path) -> None:
    dawgz.set_dawgz_dir(tmp_path / ".dawgz")


@pytest.fixture(autouse=True)
def mock_shutil_which() -> Generator[MagicMock]:
    with patch("shutil.which", return_value="/usr/bin/sbatch") as m:
        yield m


@pytest.fixture(autouse=True)
def squeue() -> dict[str, str]:
    return {}


@pytest.fixture(autouse=True)
def mock_subprocess_run(squeue: dict) -> Generator[MagicMock]:
    def run(cmd: list[str], **ignore) -> subprocess.CompletedProcess[str]:
        if cmd[0] == "sbatch":
            with open(cmd[-1]) as shfile:
                match = re.search(r"--array=(\d+)-(\d+)", shfile.read())
                if match:
                    array = range(int(match.group(1)), int(match.group(2)) + 1)
                else:
                    array = None

            jobid = f"{len(squeue):03d}"

            if array is None:
                squeue[jobid] = "PENDING"
            else:
                for i in array:
                    squeue[f"{jobid}_{i}"] = "PENDING"

            return subprocess.CompletedProcess(
                args=cmd, returncode=0, stdout=f"{jobid}\n", stderr=""
            )
        elif cmd[0] == "scancel":
            prefixes = cmd[2:]

            for jobid in squeue:
                if any(jobid.startswith(prefix) for prefix in prefixes):
                    squeue[jobid] = "CANCELLED"

            return subprocess.CompletedProcess(
                args=cmd, returncode=0, stdout="terminating job\n", stderr=""
            )
        elif cmd[0] == "sacct":
            prefix = cmd[2]
            lines = []

            for jobid, status in squeue.items():
                if jobid.startswith(prefix):
                    lines.append(f"{jobid}|{status}")

            return subprocess.CompletedProcess(
                args=cmd, returncode=0, stdout="\n".join(lines) + "\n", stderr=""
            )
        else:
            raise NotImplementedError(f"Unknown command {cmd[0]}")

    with patch("subprocess.run", side_effect=run) as m:
        yield m


@pytest.fixture(autouse=True)
def mock_srun(tmp_path: Path) -> Generator[dict[str, str]]:
    srun = tmp_path / "srun"
    srun.write_text('#!/bin/bash\nexec "$@"\n')
    srun.chmod(0o755)

    mock_env = {**os.environ, "PATH": f"{tmp_path}:" + os.environ["PATH"]}

    with patch("os.environ", mock_env) as env:
        yield env


def sbatch(shfile: Path, run: object = subprocess.run) -> None:
    content = shfile.read_text()

    # Find output log path
    match = re.search(r"#SBATCH --output=(.+)", content)
    logfile = match.group(1)

    # Find array range
    match = re.search(r"#SBATCH --array=(\d+)-(\d+)", content)

    if match:
        array = range(int(match.group(1)), int(match.group(2)) + 1)
    else:
        array = None

    # Run
    if array is None:
        result = run(
            ["/bin/bash", str(shfile)],
            capture_output=True,
            text=True,
            env=os.environ,
        )

        with open(logfile, mode="w") as f:
            f.write(result.stdout)
    else:
        for i in array:
            result = run(
                ["/bin/bash", str(shfile)],
                capture_output=True,
                text=True,
                env={**os.environ, "SLURM_ARRAY_TASK_ID": str(i)},
            )

            with open(logfile.replace("%a", str(i)), mode="w") as f:
                f.write(result.stdout)


#########
# Tests #
#########


def test_shfile() -> None:
    def hello() -> None:
        pass

    job = dawgz.job(
        hello,
        name="hello",
        shell="/bin/zsh",
        interpreter="python3",
        env=["export FOO=bar", "module load cuda"],
        settings={
            "cpus": 4,
            "ram": "16GB",
            "partition": "gpu",
            "tasks_per_node": 8,
            "gpus_per_node": 8,
        },
    )()

    scheduler = dawgz.schedule(job, backend="slurm", quiet=True)

    tag = scheduler.tag(job)
    shfile = scheduler.path / f"{tag}.sh"
    content = shfile.read_text()

    # Shell
    assert content.startswith("#!/bin/zsh\n")

    # Name
    assert f"--job-name={tag}" in content

    # Env
    assert "export FOO=bar\n" in content
    assert "module load cuda\n" in content

    # Interpreter
    assert "srun python3 " in content

    # Settings
    assert "SBATCH --nodes=1" in content
    assert "SBATCH --ntasks-per-node=8" in content
    assert "SBATCH --cpus-per-task=4" in content
    assert "SBATCH --gpus-per-node=8" in content
    assert "SBATCH --mem=16GB" in content
    assert "SBATCH --partition=gpu" in content


@pytest.mark.parametrize("wait_mode", ["all", "any"])
@pytest.mark.parametrize("status", ["success", "failure"])
def test_dependencies(wait_mode: str, status: str) -> None:
    @dawgz.job
    def f() -> None:
        pass

    a_job = f()
    b_job = f()
    c_job = f().after(a_job, b_job, status=status).waitfor(wait_mode)

    scheduler = dawgz.schedule(c_job, backend="slurm", quiet=True)

    a_jobid = scheduler.results[a_job]
    b_jobid = scheduler.results[b_job]
    c_tag = scheduler.tag(c_job)

    shfile = scheduler.path / f"{c_tag}.sh"
    content = shfile.read_text()

    sep = "," if wait_mode == "all" else "?"
    after = "afterok" if status == "success" else "afternotok"

    assert f"--dependency={after}:{a_jobid}{sep}{after}:{b_jobid}" in content


def test_dependency_submission_failure() -> None:
    @dawgz.job
    def f() -> None:
        pass

    a_job = f()
    b_job = f().after(a_job)

    def failing_sbatch(cmd: list[str], **ignore) -> subprocess.CompletedProcess[str]:
        if cmd[0] == "sbatch":
            raise subprocess.CalledProcessError(cmd=cmd, returncode=1, stderr="submission error")
        return subprocess.CompletedProcess(args=cmd, returncode=0, stdout="", stderr="")

    with patch("subprocess.run", side_effect=failing_sbatch):
        scheduler = dawgz.schedule(b_job, backend="slurm", quiet=True)

    assert "JobSubmissionError" in scheduler.traces[a_job]
    assert "DependencyNeverSatisfiedError" in scheduler.traces[b_job]


def test_single_job() -> None:
    def print_env() -> None:
        print(os.environ["DAWGZ_TEST_VAR"])

    job = dawgz.job(
        print_env,
        interpreter=sys.executable,
        env=["export DAWGZ_TEST_VAR=hello_world"],
    )()

    scheduler = dawgz.schedule(job, backend="slurm", quiet=True)

    tag = scheduler.tag(job)
    shfile = scheduler.path / f"{tag}.sh"

    sbatch(shfile)

    assert scheduler.output(job) == "hello_world\n"


def test_job_array() -> None:
    def print_item(x: object) -> None:
        print(x)

    xs = ["a", "b", "c"]
    jobs = [
        dawgz.job(
            print_item,
            interpreter=sys.executable,
        )(x)
        for x in xs
    ]
    array = dawgz.array(*jobs)

    scheduler = dawgz.schedule(array, backend="slurm", quiet=True)

    tag = scheduler.tag(array)
    shfile = scheduler.path / f"{tag}.sh"

    sbatch(shfile)

    for i, x in enumerate(xs):
        assert scheduler.output(array, i) == f"{x}\n"


def test_cancel(squeue: dict) -> None:
    @dawgz.job
    def f() -> None:
        pass

    job = f()

    scheduler = dawgz.schedule(job, backend="slurm", quiet=True)
    assert scheduler.state(job) == "PENDING"

    SlurmScheduler.sacct.cache_clear()

    scheduler.cancel(job)
    assert scheduler.state(job) == "CANCELLED"



r"""Constants"""

import os

from pathlib import Path

DAWGZ_DIR: str | Path | None = None


def get_dawgz_dir() -> Path:
    if DAWGZ_DIR is None:
        dawgz_dir = os.environ.get("DAWGZ_DIR", ".dawgz")
    else:
        dawgz_dir = DAWGZ_DIR

    return Path(dawgz_dir).expanduser().resolve()


def set_dawgz_dir(value: str | Path | None) -> Path:
    global DAWGZ_DIR
    DAWGZ_DIR = value

    return get_dawgz_dir()



r"""Module's main"""

import argparse
import csv

from tabulate import tabulate

from dawgz import Scheduler, get_dawgz_dir


def list_workflows() -> list[list[str]]:
    record = get_dawgz_dir() / "workflows.csv"

    if record.exists():
        with open(record) as f:
            return list(csv.reader(f))
    else:
        return []


def table(
    workflow: int | None = None,
    job: int | None = None,
    i: int | None = None,
) -> None:
    workflows = list_workflows()

    if workflow is None:
        headers = ("Name", "ID", "Date", "Backend", "Jobs", "Errors")
        table = tabulate(workflows, headers, showindex=True)
    else:
        row = workflows[workflow]
        uid = row[1]
        scheduler = Scheduler.load(get_dawgz_dir() / uid)

        if job is None:
            table = scheduler.report()
        else:
            jobs = list(scheduler.order)
            job = jobs[job]

            table = scheduler.report(job, i)

    try:
        print(table)
    except BrokenPipeError:
        pass


def cancel(
    workflow: int,
    job: int | None = None,
    i: int | None = None,
) -> None:
    workflows = list_workflows()

    row = workflows[workflow]
    uuid = row[1]
    scheduler = Scheduler.load(get_dawgz_dir() / uuid)

    if job is None:
        message = scheduler.cancel()
    else:
        jobs = list(scheduler.order)
        job = jobs[job]

        message = scheduler.cancel(job, i)

    if message:
        print(message)


def main() -> None:
    # Parser
    parser = argparse.ArgumentParser(description="DAWGZ's CLI")

    parser.add_argument("workflow", default=None, nargs="?", type=int, help="workflow index")
    parser.add_argument("job", default=None, nargs="?", type=int, help="job index")
    parser.add_argument("i", default=None, nargs="?", type=int, help="job array index")

    parser.add_argument("-c", "--cancel", default=False, action="store_true")

    args = parser.parse_args()

    # Action
    if args.cancel:
        cancel(args.workflow, args.job, args.i)
    else:
        table(args.workflow, args.job, args.i)


if __name__ == "__main__":
    main()
